module Forward
  module Command
    class Tunnel < Base
      BASIC_AUTH_REGEX       = /\A[^\s:]+:[^\s:]+\z/i.freeze
      CNAME_REGEX            = /\A[a-z0-9]+(?:[\-\.]{1}[a-z0-9]+)*\.[a-z]{2,5}\z/i.freeze
      SUBDOMAIN_PREFIX_REGEX = /\A[a-z0-9]{1}[a-z0-9\-]+\z/i.freeze
      STATIC_HOST            = '127.0.0.1'.freeze
      STATIC_PORT            = 32645.freeze
      DEFAULT_OPTIONS        = {
        host: 'localhost',
        port: 80
      }

      def start
        print_help_and_exit! if !forwardfile? && @args.empty?

        config.create_or_load
        @options.merge!(DEFAULT_OPTIONS)
        parse_forwardfile
        parse_forwarded

        @options[:subdomain_prefix] = @args[1] if @args.length > 1
        @options[:no_auth]          = @options.fetch(:'no-auth', nil)

        validate :port, :auth, :cname, :subdomain_prefix

        client do
          authenticate_user do
            start_static_server if @options[:static_path]
            open_tunnel
          end
        end
      end

      private

      def authenticate_user(&block)
        return block.call if logged_in?

        email, password = ask_for_credentials(email)

        logger.debug("[API] authenticating user: `#{email}:#{password.gsub(/./, 'x')}'")

        API::User.authenticate(email, password) do |subdomain, token|
          config.add_account(subdomain, token)
          block.call
        end
      end

      def open_tunnel
        API::Tunnel.create(@options) do |attributes|
          Forward.tunnel = Forward::Tunnel.new(@options.merge(attributes))
        end
      end

      def start_static_server
        require 'thin'
        require 'forward/static/app'
        static_path          = @options[:static_path]
        Thin::Logging.silent = true
        logger.debug "[static] starting server on port: #{STATIC_PORT}"
        Thin::Server.start(STATIC_HOST, STATIC_PORT) { run Static::App.new(static_path) }
      end

      def forwardfile?
        @has_forwardfile ||= File.exist?(forwardfile_path)
      end

      # Returns a String file path for PWD/Forwardfile
      def forwardfile_path
        @forwardfile_path ||= File.join(Dir.pwd, 'Forwardfile')
      end

      def parse_forwardfile
        return unless forwardfile?

        options = YAML.load_file(forwardfile_path)

        raise CLIError unless options.kind_of?(Hash)

        logger.debug "[CLI] read Forwardfile `#{forwardfile_path}' as: `#{options.inspect}'"

        @options.merge!(options.symbolize_keys)
      rescue ArgumentError, SyntaxError, CLIError
        exit_with_error "Unable to parse Forwardfile at `#{forwardfile_path}'"
      end

      # Parses the arguments to determine if we're forwarding a port or host
      # and validates the port or host and updates @options if valid.
      #
      # arg - A String representing the port or host.
      #
      # Returns a Hash containing the forwarded host or port
      def parse_forwarded
        return if @args.empty?

        forwarded = @args[0]
        logger.debug "[CLI] forwarded: `#{forwarded}'"

        if Dir.exist?(forwarded)
          @options[:static_path] = forwarded
          @options[:host]        = STATIC_HOST
          @options[:port]        = STATIC_PORT
          return
        end

        case forwarded
        when /\A\d{1,5}\z/
          @options[:port] = forwarded.to_i
        when /\A[-a-z0-9\.\-]+\z/i
          @options[:host] = forwarded
        when /\A[-a-z0-9\.\-]+:\d{1,5}\z/i
          host, port = forwarded.split(':')
          port       = port.to_i

          @options[:host] = host
          @options[:port] = port
        end
      end

      # Checks to make sure the port being set is a number between 1 and 65535
      # and exits with an error message if it's not.
      #
      # port - port number Integer
      def validate_port
        port = @options[:port]

        logger.debug "[CLI] validating port: `#{port}'"
        unless port.between?(1, 65535)
          raise CLIError, "Invalid Port: #{port} is an invalid port number"
        end
      end

      def validate_auth
        auth = @options[:auth]

        return if auth.nil?

        logger.debug "[CLI] validating auth: `#{auth}'"
        if auth =~ BASIC_AUTH_REGEX
          @options[:username], @options[:password] = @options[:auth].split(':')
        else
          raise CLIError, "`#{auth}' isn't a valid username:password pair"
        end
      end

      # Checks to make sure the cname is in the correct format and exits with an
      # error message if it isn't.
      #
      # cname - cname String
      def validate_cname
        cname = @options[:cname]

        return if cname.nil?

        logger.debug "[CLI] validating CNAME: `#{cname}'"
        unless cname =~ CNAME_REGEX
          raise CLIError, "`#{cname}' is an invalid domain format"
        end
      end

      # Checks to make sure the subdomain prefix is in the correct format
      # and exits with an error message if it isn't.
      #
      # prefix - subdomain prefix String
      def validate_subdomain_prefix
        prefix = @options[:subdomain_prefix]

        return if prefix.nil?

        logger.debug "[CLI] validating subdomain prefix: `#{prefix}'"
        unless prefix =~ SUBDOMAIN_PREFIX_REGEX
          raise CLIError, "`#{prefix}' is an invalid subdomain prefix format"
        end
      end
    end
  end
end
