module Forward
  module Config
    extend self
    DEFAULTS = {
      accounts: {},
      default_account: nil,
      auto_copy: true,
      auto_open: false
    }.freeze

    attr_accessor :accounts
    attr_accessor :default_account
    attr_accessor :auto_copy
    attr_accessor :auto_open

    def set_defaults!
      DEFAULTS.keys.each { |setting| set_default!(setting) }
    end

    def set_default!(setting)
      value = DEFAULTS[setting.to_sym]
      value = value.dup if value.is_a?(Hash) || value.is_a?(Array)
      self.send("#{setting}=", value)

      value
    end

    def set_default_account(subdomain)
      self.default_account = subdomain

      write
    end

    def add_account(subdomain, api_token)
      accounts[subdomain.to_sym] = api_token
      self.default_account       = subdomain

      write
    end

    def remove_account(subdomain)
      accounts.delete(subdomain.to_sym)
      self.default_account = accounts.empty? ? nil : accounts.keys.first

      write
    end

    %w[set_default_account add_account remove_account].each do |name|
      module_eval "def #{name}!(*args); #{name}(*args); write; end"
    end

    def api_key
      accounts[default_account.to_sym]
    end

    def auto_open?
      auto_open == true
    end

    def auto_copy?
      auto_copy == true
    end

    # Updates an existing Config object.
    #
    # Returns the updated Config object.
    def update(attributes)
      Forward.logger.debug('[config] updating')
      attributes.each do |key, value|
        self.send(:"#{key}=", value)
      end

      self
    end

    # Converts a Config object to a Hash.
    #
    # Returns a Hash representation of the object
    def to_hash
      Hash[instance_variables.map { |var| [var[1..-1].to_sym, instance_variable_get(var)] }]
    end

    # Write the current config data to `config_path', and the current
    # private_key to `key_path'.
    #
    # Returns the Config object.
    def write
      Forward.logger.debug('[config] writing')

      File.open(config_path, 'w') { |f| f.write(YAML.dump(to_hash)) }

      self
    rescue Exception => e
      Forward.logger.fatal "#{e.message}"
      raise ConfigError, 'Unable to write config file'
    end

    # Returns the location of the forward config file
    # based on the host os.
    #
    # Returns the String path.
    def config_path
      File.join(ENV['HOME'], '.forwardrc')
    end

    # Checks to see if a .forward config file exists.
    #
    # Returns true or false based on the existence of the config file.
    def present?
      File.exist? config_path
    end

    # Create a config file if it doesn't exist, load one if it does.
    #
    # Returns the resulting Config object.
    def create_or_load
      if present?
        load
      else
        set_defaults!
        write
      end
    end

    # It initializes a new Config instance, updates it with the config values
    # from the config file, and raises an error if there's not a config or if
    # the config options aren't valid.
    #
    # Returns the Config object.
    def load
      Forward.logger.debug('[config] loading')
      raise ConfigError, "Unable to find a forward config file at `#{config_path}'" unless present?

      update(YAML.load_file(config_path).symbolize_keys)
    end

  end
end
