require 'erb'

module Forward
  module Static

    class TemplateContext
      FILESIZE_FORMAT = [
        ['%.1fT', 1 << 40],
        ['%.1fG', 1 << 30],
        ['%.1fM', 1 << 20],
        ['%.1fK', 1 << 10],
      ]


      def initialize(root, path_info)
        @root         = root
        @project_name = File.basename(root)
        @path_info    = path_info
        @path         = File.join(root, path_info)
      end

      def path_components
        @path_info.split('/')[1..-1]
      end

      def files
        _files = []

        Dir["#{@path}*"].sort.each do |path|
          _stat = stat(path)
          next if _stat.nil?

          basename = File.basename(path)
          url      = path.sub(@root, '')
          ext      = File.extname(path)
          size     = _stat.size
          type     = _stat.directory? ? 'directory' : Rack::Mime.mime_type(ext)
          size     = _stat.directory? ? '-' : filesize_format(size)

          if _stat.directory?
            url      << '/'
            basename << '/'
          end

          _files << {
            url: url,
            basename: basename,
            size: size,
            type: type,
          }
        end

        _files
      end

      def get_binding
        binding
      end

      private

      def stat(path)
        File.stat(path)
      rescue Errno::ENOENT, Errno::ELOOP
        nil
      end

      def filesize_format(int)
        FILESIZE_FORMAT.each do |format, size|
          return format % (int.to_f / size) if int >= size
        end

        int.to_s + 'B'
      end
    end

    class App
      attr_reader :files
      attr_accessor :root
      attr_accessor :path

      def initialize(root, app = nil)
        @root            = File.expand_path(root)
        @app             = app || Rack::File.new(@root)
        @listing_view    = File.read(File.expand_path('../directory.erb', __FILE__))
      end

      def call(env)
        dup._call(env)
      end

      def _call(env)
        @env       = env
        @path_info = Rack::Utils.unescape(env['PATH_INFO'])

        if forbidden?
          render_404
        else
          @path = File.join(@root, @path_info)
          process
        end
      end

      private

      def forbidden?
        @path_info =~ /\.\./
      end

      def render_directory_listing
        context = TemplateContext.new(@root, @path_info)
        body    = ERB.new(@listing_view).result(context.get_binding)

        [ 200, {'Content-Type' => 'text/html; charset=utf-8'}, [body] ]
      end

      def has_index?
        File.exist?(File.join(@path, 'index.html'))
      end

      def process
        return render_404 unless File.exist?(@path)

        @stat = File.stat(@path)

        raise Errno::ENOENT, 'No such file or directory' unless @stat.readable?

        if @stat.directory?
          return render_404 unless @path.end_with?('/')
          return render_directory_listing unless has_index?

          @env['PATH_INFO'] << 'index.html'
        end

        Rack::File.new(@root).call(@env)

      rescue Errno::ENOENT, Errno::ELOOP => e
        Forward.logger.debug e
        Forward.logger.debug e.message
        render_404
      end


      def render_404
        body    = "Not Found: #{@path_info}\n"
        headers = {
          "Content-Type"   => "text/plain",
          "Content-Length" => Rack::Utils.bytesize(body).to_s,
          "X-Cascade"      => "pass"
        }

        [404, headers, [body]]
      end

    end
  end
end
