module Forward
  class Tunnel
    include Common

    # The Tunnel resource ID
    attr_reader :id
    # The domain for the Tunnel
    attr_reader :subdomain
    # The CNAME for the Tunnel
    attr_reader :cname
    # The host to forward requests to
    attr_reader :host
    # The port to forward requests to
    attr_reader :port
    # Authentication for tunnel
    attr_reader :username
    attr_reader :password
    attr_reader :no_auth
    # The public hostname/subdomain url for the tunnel
    attr_reader :url
    # The tunneler host
    attr_reader :tunneler
    # The timeout
    attr_reader :timeout

    attr_reader :socket
    attr_reader :socket_url
    attr_reader :requests

    attr_accessor :last_active_at

    # Initializes a Tunnel instance for the Client and requests a tunnel from
    # API.
    #
    # client - The Client instance.
    def initialize(attributes = {})
      logger.debug "[tunnel] initializing with: #{attributes.inspect}"
      @attributes       = attributes
      @id               = attributes[:id]
      @host             = attributes[:vhost]
      @port             = attributes[:hostport]
      @subdomain_prefix = attributes[:subdomain_prefix]
      @username         = attributes[:username]
      @password         = attributes[:password]
      @no_auth          = attributes[:no_auth]
      @static_path      = attributes[:static_path]
      @cname            = attributes[:cname]
      @timeout          = attributes[:timeout]
      @tunneler         = attributes[:tunneler]
      @url              = attributes[:url]
      @requests         = {}
      @open             = false
      @socket           = Socket.new(self)
    end

    def ready!
      logger.debug '[tunnel] opened successfully'
      @open = true

      copy_url_to_clipboard
      open_url_in_browser
      display_ready_message
    end

    def open?
      @open
    end

    def destroy(&block)
      API::Tunnel.destroy(id, &block)
    end

    def authority
      @authority ||= begin
        _authority = "#{host}"
        _authority << ":#{port}" unless port == 80

        _authority
      end
    end

    def track_activity!
      self.last_active_at = Time.now.to_i
    end

    private

    def display_ready_message
      source  = static? ? "Directory `#{File.expand_path(@static_path).split('/').last}'" : authority
      message = "#{source} is now available at: #{HighLine.color(url, :underline)}"

      message << " and #{HighLine.color("http://#{cname}", :underline)}" if cname

      puts "#{message}\n\nCtrl-C to stop forwarding"
    end

    def copy_url_to_clipboard
      return unless config.auto_copy?

      if windows?
        begin
          require 'ffi'
        rescue LoadError
          puts "The FFI gem is required to copy your url to the clipboard, you can install it with `gem install ffi'"
          return
        end
      end

      Clipboard.copy(url)
    end

    def open_url_in_browser
      return unless config.auto_open?

      case os
      when :windows
        %x[start #{url}]
      when :osx
        %x[open #{url}]
      when :unix
        %x[xdg-open #{url}]
      end
    end

    def static?
      !@static_path.nil?
    end

  end
end
